



if not exist %LIBRARY_LIB%\\astra.lib exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_BIN%\\astra.dll exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
tests\\astratest
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
